@@ -11,6 +11,10 @@
 #include <fcntl.h>
 #endif
 
+#ifdef __linux__
+#include <sys/ptrace.h>
+#endif
+
 /* We write 2^1 + 2^2 ... + 2^10 or 2047 copies of "Hello World!\n"
  * ultimately
  */
@@ -757,11 +761,15 @@ on_communicate_complete (GObject               *proc,
     {
       if (data->is_utf8)
         {
+          g_assert_nonnull (stdout_str);
           stdout_data = (guint8*)stdout_str;
           stdout_len = strlen (stdout_str);
         }
       else
-        stdout_data = g_bytes_get_data (stdout_bytes, &stdout_len);
+        {
+          g_assert_nonnull (stdout_bytes);
+          stdout_data = g_bytes_get_data (stdout_bytes, &stdout_len);
+        }
 
       g_assert_cmpmem (stdout_data, stdout_len, "# hello world" LINEEND, 13 + strlen (LINEEND));
     }
@@ -1284,14 +1292,12 @@ test_communicate_utf8_invalid (void)
   g_object_unref (proc);
 }
 
-static gboolean
+static void
 send_terminate (gpointer   user_data)
 {
   GSubprocess *proc = user_data;
 
   g_subprocess_force_exit (proc);
-
-  return FALSE;
 }
 
 static void
@@ -1337,7 +1343,7 @@ test_terminate (void)
 
   g_subprocess_wait_async (proc, NULL, on_request_quit_exited, loop);
 
-  g_timeout_add_seconds (3, send_terminate, proc);
+  g_timeout_add_seconds_once (3, send_terminate, proc);
 
   g_main_loop_run (loop);
 
@@ -1346,14 +1352,12 @@ test_terminate (void)
 }
 
 #ifdef G_OS_UNIX
-static gboolean
+static void
 send_signal (gpointer user_data)
 {
   GSubprocess *proc = user_data;
 
   g_subprocess_send_signal (proc, SIGKILL);
-
-  return FALSE;
 }
 
 static void
@@ -1374,7 +1378,7 @@ test_signal (void)
 
   g_subprocess_wait_async (proc, NULL, on_request_quit_exited, loop);
 
-  g_timeout_add_seconds (3, send_signal, proc);
+  g_timeout_add_seconds_once (3, send_signal, proc);
 
   g_main_loop_run (loop);
 
@@ -1746,9 +1750,9 @@ do_test_pass_fd (GSubprocessFlags     flags,
   char *basic_fd_str;
   char *needdup_fd_str;
 
-  g_unix_open_pipe (basic_pipefds, FD_CLOEXEC, error);
+  g_unix_open_pipe (basic_pipefds, O_CLOEXEC, error);
   g_assert_no_error (local_error);
-  g_unix_open_pipe (needdup_pipefds, FD_CLOEXEC, error);
+  g_unix_open_pipe (needdup_pipefds, O_CLOEXEC, error);
   g_assert_no_error (local_error);
 
   basic_fd_str = g_strdup_printf ("%d", basic_pipefds[1]);
@@ -1850,10 +1854,10 @@ do_test_fd_conflation (GSubprocessFlags     flags,
       return;
     }
 
-  g_unix_open_pipe (unused_pipefds, FD_CLOEXEC, &error);
+  g_unix_open_pipe (unused_pipefds, O_CLOEXEC, &error);
   g_assert_no_error (error);
 
-  g_unix_open_pipe (pipefds, FD_CLOEXEC, &error);
+  g_unix_open_pipe (pipefds, O_CLOEXEC, &error);
   g_assert_no_error (error);
 
   /* The fds should be sequential since we are in a new process. */
@@ -1989,8 +1993,118 @@ test_fd_conflation_child_err_report_fd (void)
   do_test_fd_conflation (G_SUBPROCESS_FLAGS_NONE, empty_child_setup, TRUE);
 }
 
+#ifdef __linux__
+
+/* Handle ptrace events on @main_child, and assert that when it exits, it does
+ * so with status %EXIT_SUCCESS, rather than signalling. Other than that, this
+ * just calls %PTRACE_CONT for all trace events. */
+static void
+trace_children (pid_t main_child)
+{
+  int wstatus;
+
+  g_assert_no_errno (waitpid (main_child, &wstatus, 0));
+  g_assert_no_errno (ptrace (PTRACE_SETOPTIONS, main_child, NULL,
+                             (PTRACE_O_TRACEFORK |
+                              PTRACE_O_EXITKILL |
+                              PTRACE_O_TRACEVFORK |
+                              PTRACE_O_TRACECLONE |
+                              PTRACE_O_TRACEEXEC)));
+  g_assert_no_errno (ptrace (PTRACE_CONT, main_child, NULL, 0));
+
+  while (TRUE)
+    {
+      pid_t pid;
+      int wstatus;
+      int stop_signum;
+      int ptrace_event;
+
+      pid = waitpid (-1, &wstatus, 0);
+      if (pid == -1 && errno == ECHILD)
+        break;
+
+      g_assert_cmpint (errno, ==, 0);
+      g_assert_cmpint (pid, >=, 0);
+
+      if (WIFSTOPPED (wstatus))
+        stop_signum = WSTOPSIG (wstatus);
+      else
+        stop_signum = 0;
+
+      switch (stop_signum)
+        {
+        case SIGTRAP:
+          ptrace_event = (wstatus >> 16) & 0xffff;
+          switch (ptrace_event)
+            {
+            case 0:
+              g_assert_no_errno (ptrace (PTRACE_CONT, pid, NULL, stop_signum));
+              break;
+            default:
+              g_assert_no_errno (ptrace (PTRACE_CONT, pid, NULL, 0));
+              break;
+            }
+          break;
+        case SIGSTOP:
+          g_assert_no_errno (ptrace (PTRACE_CONT, pid, NULL, 0));
+          break;
+        default:
+          if (!WIFEXITED (wstatus) && !WIFSIGNALED (wstatus))
+            g_assert_no_errno (ptrace (PTRACE_CONT, pid, NULL, stop_signum));
+          break;
+        }
+
+      if (pid == main_child)
+        {
+          g_assert_false (WIFSIGNALED (wstatus));
+          if (WIFEXITED (wstatus))
+            {
+              g_assert_cmpint (WEXITSTATUS (wstatus), ==, EXIT_SUCCESS);
+              break;
+            }
+        }
+    }
+}
+
+#endif  /* __linux__ */
+
+static void
+test_exit_status_trapped (void)
+{
+#ifdef __linux__
+  GPtrArray *args = NULL;
+  pid_t test_child;
 #endif
 
+  g_test_summary ("Test that exit status is reported correctly for ptrace()d child processes");
+  g_test_bug ("https://gitlab.gnome.org/GNOME/glib/-/merge_requests/3433");
+
+#ifdef __linux__
+  /* Call fork() directly here, rather than using #GSubprocess, so that we can
+   * safely call waitpid() on it ourselves without interfering with the internals
+   * of #GSubprocess.
+   * See https://gitlab.gnome.org/GNOME/glib/-/merge_requests/3433#note_1749055 */
+  args = get_test_subprocess_args ("sleep-and-kill", NULL);
+  test_child = fork ();
+  if (test_child == 0)
+    {
+      /* Between fork() and exec() we can only call async-signal-safe functions. */
+      if (ptrace (PTRACE_TRACEME, 0, NULL, NULL) < 0)
+        abort ();
+
+      g_assert_no_errno (execvp (args->pdata[0], (char * const *) args->pdata));
+    }
+
+  trace_children (test_child);
+
+  g_clear_pointer (&args, g_ptr_array_unref);
+#else
+  g_test_skip ("ptrace() support for this test is only tested on Linux");
+#endif
+}
+
+#endif  /* G_OS_UNIX */
+
 static void
 test_launcher_environment (void)
 {
@@ -2133,6 +2247,7 @@ main (int argc, char **argv)
   g_test_add_func ("/gsubprocess/fd-conflation/empty-child-setup", test_fd_conflation_empty_child_setup);
   g_test_add_func ("/gsubprocess/fd-conflation/inherit-fds", test_fd_conflation_inherit_fds);
   g_test_add_func ("/gsubprocess/fd-conflation/child-err-report-fd", test_fd_conflation_child_err_report_fd);
+  g_test_add_func ("/gsubprocess/exit-status/trapped", test_exit_status_trapped);
 #endif
   g_test_add_func ("/gsubprocess/launcher-environment", test_launcher_environment);
 
